# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/dot15d4/dot15d4.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#whad/protocol/dot15d4/dot15d4.proto\x12\x07\x64ot15d4\"P\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12*\n\x0c\x61\x64\x64ress_type\x18\x02 \x01(\x0e\x32\x14.dot15d4.AddressType\"\x1b\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"%\n\x12\x45nergyDetectionCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\'\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0b\n\x03pdu\x18\x02 \x01(\x0c\"7\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0b\n\x03pdu\x18\x02 \x01(\x0c\x12\x0b\n\x03\x66\x63s\x18\x03 \x01(\r\"\x1f\n\x0c\x45ndDeviceCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\x1c\n\tRouterCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"!\n\x0e\x43oordinatorCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\";\n\x11ManInTheMiddleCmd\x12&\n\x04role\x18\x01 \x01(\x0e\x32\x18.dot15d4.Dot15d4MitmRole\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\":\n\x15\x45nergyDetectionSample\x12\x0e\n\x06sample\x18\x01 \x01(\r\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\xc3\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x66\x63s_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x0b\n\x03pdu\x18\x05 \x01(\x0c\x12\x0b\n\x03\x66\x63s\x18\x06 \x01(\r\x12\x10\n\x03lqi\x18\x07 \x01(\rH\x03\x88\x01\x01\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_fcs_validityB\x06\n\x04_lqi\"\xb3\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x66\x63s_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x0b\n\x03pdu\x18\x05 \x01(\x0c\x12\x10\n\x03lqi\x18\x06 \x01(\rH\x03\x88\x01\x01\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_fcs_validityB\x06\n\x04_lqi\"\x9d\x05\n\x07Message\x12\x33\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x1a.dot15d4.SetNodeAddressCmdH\x00\x12\"\n\x05sniff\x18\x02 \x01(\x0b\x32\x11.dot15d4.SniffCmdH\x00\x12\x1e\n\x03jam\x18\x03 \x01(\x0b\x32\x0f.dot15d4.JamCmdH\x00\x12)\n\x02\x65\x64\x18\x04 \x01(\x0b\x32\x1b.dot15d4.EnergyDetectionCmdH\x00\x12 \n\x04send\x18\x05 \x01(\x0b\x32\x10.dot15d4.SendCmdH\x00\x12\'\n\x08send_raw\x18\x06 \x01(\x0b\x32\x13.dot15d4.SendRawCmdH\x00\x12+\n\nend_device\x18\x07 \x01(\x0b\x32\x15.dot15d4.EndDeviceCmdH\x00\x12$\n\x06router\x18\x08 \x01(\x0b\x32\x12.dot15d4.RouterCmdH\x00\x12.\n\x0b\x63oordinator\x18\t \x01(\x0b\x32\x17.dot15d4.CoordinatorCmdH\x00\x12\"\n\x05start\x18\n \x01(\x0b\x32\x11.dot15d4.StartCmdH\x00\x12 \n\x04stop\x18\x0b \x01(\x0b\x32\x10.dot15d4.StopCmdH\x00\x12*\n\x04mitm\x18\x0c \x01(\x0b\x32\x1a.dot15d4.ManInTheMiddleCmdH\x00\x12!\n\x06jammed\x18\r \x01(\x0b\x32\x0f.dot15d4.JammedH\x00\x12\x33\n\ted_sample\x18\x0e \x01(\x0b\x32\x1e.dot15d4.EnergyDetectionSampleH\x00\x12*\n\x07raw_pdu\x18\x0f \x01(\x0b\x32\x17.dot15d4.RawPduReceivedH\x00\x12#\n\x03pdu\x18\x10 \x01(\x0b\x32\x14.dot15d4.PduReceivedH\x00\x42\x05\n\x03msg*\xc5\x01\n\x0e\x44ot15d4Command\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x13\n\x0f\x45nergyDetection\x10\x03\x12\x08\n\x04Send\x10\x04\x12\x0b\n\x07SendRaw\x10\x05\x12\x11\n\rEndDeviceMode\x10\x06\x12\x13\n\x0f\x43oordinatorMode\x10\x07\x12\x0e\n\nRouterMode\x10\x08\x12\t\n\x05Start\x10\t\x12\x08\n\x04Stop\x10\n\x12\x12\n\x0eManInTheMiddle\x10\x0b*5\n\x0f\x44ot15d4MitmRole\x12\x13\n\x0fREACTIVE_JAMMER\x10\x00\x12\r\n\tCORRECTOR\x10\x01*&\n\x0b\x41\x64\x64ressType\x12\t\n\x05SHORT\x10\x00\x12\x0c\n\x08\x45XTENDED\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.dot15d4.dot15d4_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DOT15D4COMMAND._serialized_start=1647
  _DOT15D4COMMAND._serialized_end=1844
  _DOT15D4MITMROLE._serialized_start=1846
  _DOT15D4MITMROLE._serialized_end=1899
  _ADDRESSTYPE._serialized_start=1901
  _ADDRESSTYPE._serialized_end=1939
  _SETNODEADDRESSCMD._serialized_start=48
  _SETNODEADDRESSCMD._serialized_end=128
  _SNIFFCMD._serialized_start=130
  _SNIFFCMD._serialized_end=157
  _ENERGYDETECTIONCMD._serialized_start=159
  _ENERGYDETECTIONCMD._serialized_end=196
  _JAMCMD._serialized_start=198
  _JAMCMD._serialized_end=223
  _SENDCMD._serialized_start=225
  _SENDCMD._serialized_end=264
  _SENDRAWCMD._serialized_start=266
  _SENDRAWCMD._serialized_end=321
  _ENDDEVICECMD._serialized_start=323
  _ENDDEVICECMD._serialized_end=354
  _ROUTERCMD._serialized_start=356
  _ROUTERCMD._serialized_end=384
  _COORDINATORCMD._serialized_start=386
  _COORDINATORCMD._serialized_end=419
  _STARTCMD._serialized_start=421
  _STARTCMD._serialized_end=431
  _STOPCMD._serialized_start=433
  _STOPCMD._serialized_end=442
  _MANINTHEMIDDLECMD._serialized_start=444
  _MANINTHEMIDDLECMD._serialized_end=503
  _JAMMED._serialized_start=505
  _JAMMED._serialized_end=532
  _ENERGYDETECTIONSAMPLE._serialized_start=534
  _ENERGYDETECTIONSAMPLE._serialized_end=592
  _RAWPDURECEIVED._serialized_start=595
  _RAWPDURECEIVED._serialized_end=790
  _PDURECEIVED._serialized_start=793
  _PDURECEIVED._serialized_end=972
  _MESSAGE._serialized_start=975
  _MESSAGE._serialized_end=1644
# @@protoc_insertion_point(module_scope)
