# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/phy/phy.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/phy/phy.proto\x12\x03phy\"\"\n\x13SetASKModulationCmd\x12\x0b\n\x03ook\x18\x01 \x01(\x08\"(\n\x13SetFSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\")\n\x14Set4FSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\")\n\x14SetGFSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\"(\n\x13SetMSKModulationCmd\x12\x11\n\tdeviation\x18\x01 \x01(\r\"\x16\n\x14SetBPSKModulationCmd\"+\n\x14SetQPSKModulationCmd\x12\x13\n\x0boffset_qpsk\x18\x01 \x01(\x08\"\xde\x01\n\x14SetLoRaModulationCmd\x12\x11\n\tbandwidth\x18\x01 \x01(\r\x12\x32\n\x10spreading_factor\x18\x02 \x01(\x0e\x32\x18.phy.LoRaSpreadingFactor\x12(\n\x0b\x63oding_rate\x18\x03 \x01(\x0e\x32\x13.phy.LoRaCodingRate\x12\x17\n\x0fpreamble_length\x18\x04 \x01(\r\x12\x12\n\nenable_crc\x18\x05 \x01(\x08\x12\x15\n\rexplicit_mode\x18\x06 \x01(\x08\x12\x11\n\tinvert_iq\x18\x07 \x01(\x08\"\x1c\n\x1aGetSupportedFrequenciesCmd\"$\n\x0fSetFrequencyCmd\x12\x11\n\tfrequency\x18\x01 \x01(\r\"\x1e\n\x0eSetDataRateCmd\x12\x0c\n\x04rate\x18\x01 \x01(\r\"7\n\x10SetEndiannessCmd\x12#\n\nendianness\x18\x01 \x01(\x0e\x32\x0f.phy.Endianness\"/\n\rSetTXPowerCmd\x12\x1e\n\x08tx_power\x18\x01 \x01(\x0e\x32\x0c.phy.TXPower\"\'\n\x10SetPacketSizeCmd\x12\x13\n\x0bpacket_size\x18\x01 \x01(\r\"#\n\x0eSetSyncWordCmd\x12\x11\n\tsync_word\x18\x01 \x01(\x0c\"0\n\x08SniffCmd\x12\x16\n\tiq_stream\x18\x01 \x01(\x08H\x00\x88\x01\x01\x42\x0c\n\n_iq_stream\"\x19\n\x07SendCmd\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\"\x1c\n\nSendRawCmd\x12\x0e\n\x02iq\x18\x01 \x03(\x05\x42\x02\x10\x01\"4\n\x0fScheduleSendCmd\x12\x0e\n\x06packet\x18\x01 \x01(\x0c\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"<\n\x12SchedulePacketResp\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x11\n\x04\x66ull\x18\x02 \x01(\x08H\x00\x88\x01\x01\x42\x07\n\x05_full\" \n\x12SchedulePacketSent\x12\n\n\x02id\x18\x01 \x01(\x05\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"(\n\x06JamCmd\x12\x1e\n\x04mode\x18\x01 \x01(\x0e\x32\x10.phy.JammingMode\"\x0c\n\nMonitorCmd\"\xf2\x01\n\x0ePacketReceived\x12\x11\n\tfrequency\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x0e\n\x06packet\x18\x04 \x01(\x0c\x12\x11\n\tdeviation\x18\x05 \x01(\r\x12\x10\n\x08\x64\x61tarate\x18\x06 \x01(\r\x12\x1f\n\x06\x65ndian\x18\x07 \x01(\x0e\x32\x0f.phy.Endianness\x12#\n\nmodulation\x18\x08 \x01(\x0e\x32\x0f.phy.Modulation\x12\x10\n\x08syncword\x18\t \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestamp\"\x85\x02\n\x11RawPacketReceived\x12\x11\n\tfrequency\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x0e\n\x06packet\x18\x04 \x01(\x0c\x12\x0e\n\x02iq\x18\x05 \x03(\x05\x42\x02\x10\x01\x12\x11\n\tdeviation\x18\x06 \x01(\r\x12\x10\n\x08\x64\x61tarate\x18\x07 \x01(\r\x12\x1f\n\x06\x65ndian\x18\x08 \x01(\x0e\x32\x0f.phy.Endianness\x12#\n\nmodulation\x18\t \x01(\x0e\x32\x0f.phy.Modulation\x12\x10\n\x08syncword\x18\n \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestamp\"\x1f\n\nPacketSent\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"9\n\x10MonitoringReport\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x12\n\x06report\x18\x02 \x03(\rB\x02\x10\x01\"\x90\x01\n\x18SupportedFrequencyRanges\x12\x46\n\x10\x66requency_ranges\x18\x01 \x03(\x0b\x32,.phy.SupportedFrequencyRanges.FrequencyRange\x1a,\n\x0e\x46requencyRange\x12\r\n\x05start\x18\x01 \x01(\r\x12\x0b\n\x03\x65nd\x18\x02 \x01(\r\"\x91\n\n\x07Message\x12+\n\x07mod_ask\x18\x01 \x01(\x0b\x32\x18.phy.SetASKModulationCmdH\x00\x12+\n\x07mod_fsk\x18\x02 \x01(\x0b\x32\x18.phy.SetFSKModulationCmdH\x00\x12-\n\x08mod_gfsk\x18\x03 \x01(\x0b\x32\x19.phy.SetGFSKModulationCmdH\x00\x12-\n\x08mod_bpsk\x18\x04 \x01(\x0b\x32\x19.phy.SetBPSKModulationCmdH\x00\x12-\n\x08mod_qpsk\x18\x05 \x01(\x0b\x32\x19.phy.SetQPSKModulationCmdH\x00\x12-\n\x08mod_4fsk\x18\x06 \x01(\x0b\x32\x19.phy.Set4FSKModulationCmdH\x00\x12+\n\x07mod_msk\x18\x07 \x01(\x0b\x32\x18.phy.SetMSKModulationCmdH\x00\x12=\n\x12get_supported_freq\x18\x08 \x01(\x0b\x32\x1f.phy.GetSupportedFrequenciesCmdH\x00\x12(\n\x08set_freq\x18\t \x01(\x0b\x32\x14.phy.SetFrequencyCmdH\x00\x12\'\n\x08\x64\x61tarate\x18\n \x01(\x0b\x32\x13.phy.SetDataRateCmdH\x00\x12+\n\nendianness\x18\x0b \x01(\x0b\x32\x15.phy.SetEndiannessCmdH\x00\x12&\n\x08tx_power\x18\x0c \x01(\x0b\x32\x12.phy.SetTXPowerCmdH\x00\x12,\n\x0bpacket_size\x18\r \x01(\x0b\x32\x15.phy.SetPacketSizeCmdH\x00\x12(\n\tsync_word\x18\x0e \x01(\x0b\x32\x13.phy.SetSyncWordCmdH\x00\x12\x1e\n\x05sniff\x18\x0f \x01(\x0b\x32\r.phy.SniffCmdH\x00\x12\x1c\n\x04send\x18\x10 \x01(\x0b\x32\x0c.phy.SendCmdH\x00\x12#\n\x08send_raw\x18\x11 \x01(\x0b\x32\x0f.phy.SendRawCmdH\x00\x12\x1e\n\x05start\x18\x12 \x01(\x0b\x32\r.phy.StartCmdH\x00\x12\x1c\n\x04stop\x18\x13 \x01(\x0b\x32\x0c.phy.StopCmdH\x00\x12\x1a\n\x03jam\x18\x14 \x01(\x0b\x32\x0b.phy.JamCmdH\x00\x12\"\n\x07monitor\x18\x15 \x01(\x0b\x32\x0f.phy.MonitorCmdH\x00\x12%\n\x06packet\x18\x16 \x01(\x0b\x32\x13.phy.PacketReceivedH\x00\x12,\n\nraw_packet\x18\x17 \x01(\x0b\x32\x16.phy.RawPacketReceivedH\x00\x12\x1d\n\x06jammed\x18\x18 \x01(\x0b\x32\x0b.phy.JammedH\x00\x12/\n\x0emonitor_report\x18\x19 \x01(\x0b\x32\x15.phy.MonitoringReportH\x00\x12\x37\n\x0esupported_freq\x18\x1a \x01(\x0b\x32\x1d.phy.SupportedFrequencyRangesH\x00\x12-\n\x08mod_lora\x18\x1b \x01(\x0b\x32\x19.phy.SetLoRaModulationCmdH\x00\x12*\n\nsched_send\x18\x1c \x01(\x0b\x32\x14.phy.ScheduleSendCmdH\x00\x12\x30\n\rsched_pkt_rsp\x18\x1d \x01(\x0b\x32\x17.phy.SchedulePacketRespH\x00\x12\x31\n\x0esched_pkt_sent\x18\x1e \x01(\x0b\x32\x17.phy.SchedulePacketSentH\x00\x42\x05\n\x03msg*\xa7\x03\n\nPhyCommand\x12\x14\n\x10SetASKModulation\x10\x00\x12\x14\n\x10SetFSKModulation\x10\x01\x12\x15\n\x11SetGFSKModulation\x10\x02\x12\x15\n\x11SetBPSKModulation\x10\x03\x12\x15\n\x11SetQPSKModulation\x10\x04\x12\x15\n\x11Set4FSKModulation\x10\x05\x12\x14\n\x10SetMSKModulation\x10\x06\x12\x1b\n\x17GetSupportedFrequencies\x10\x07\x12\x10\n\x0cSetFrequency\x10\x08\x12\x0f\n\x0bSetDataRate\x10\t\x12\x11\n\rSetEndianness\x10\n\x12\x0e\n\nSetTXPower\x10\x0b\x12\x11\n\rSetPacketSize\x10\x0c\x12\x0f\n\x0bSetSyncWord\x10\r\x12\t\n\x05Sniff\x10\x0e\x12\x08\n\x04Send\x10\x0f\x12\x0b\n\x07SendRaw\x10\x10\x12\x07\n\x03Jam\x10\x11\x12\x0b\n\x07Monitor\x10\x12\x12\t\n\x05Start\x10\x13\x12\x08\n\x04Stop\x10\x14\x12\x15\n\x11SetLoRaModulation\x10\x15\x12\x10\n\x0cScheduleSend\x10\x16*!\n\nEndianness\x12\x07\n\x03\x42IG\x10\x00\x12\n\n\x06LITTLE\x10\x01*(\n\x07TXPower\x12\x07\n\x03LOW\x10\x00\x12\n\n\x06MEDIUM\x10\x01\x12\x08\n\x04HIGH\x10\x02*+\n\x0bJammingMode\x12\x0e\n\nCONTINUOUS\x10\x00\x12\x0c\n\x08REACTIVE\x10\x01*N\n\x13LoRaSpreadingFactor\x12\x07\n\x03SF7\x10\x00\x12\x07\n\x03SF8\x10\x01\x12\x07\n\x03SF9\x10\x02\x12\x08\n\x04SF10\x10\x03\x12\x08\n\x04SF11\x10\x04\x12\x08\n\x04SF12\x10\x05*8\n\x0eLoRaCodingRate\x12\x08\n\x04\x43R45\x10\x00\x12\x08\n\x04\x43R46\x10\x01\x12\x08\n\x04\x43R47\x10\x02\x12\x08\n\x04\x43R48\x10\x03*\\\n\nModulation\x12\x07\n\x03\x41SK\x10\x00\x12\x07\n\x03\x46SK\x10\x01\x12\x0b\n\x07\x46OURFSK\x10\x02\x12\x08\n\x04GFSK\x10\x03\x12\x07\n\x03MSK\x10\x04\x12\x08\n\x04\x42PSK\x10\x05\x12\x08\n\x04QPSK\x10\x06\x12\x08\n\x04LORA\x10\x07\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.phy.phy_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SENDRAWCMD.fields_by_name['iq']._options = None
  _SENDRAWCMD.fields_by_name['iq']._serialized_options = b'\020\001'
  _RAWPACKETRECEIVED.fields_by_name['iq']._options = None
  _RAWPACKETRECEIVED.fields_by_name['iq']._serialized_options = b'\020\001'
  _MONITORINGREPORT.fields_by_name['report']._options = None
  _MONITORINGREPORT.fields_by_name['report']._serialized_options = b'\020\001'
  _PHYCOMMAND._serialized_start=3234
  _PHYCOMMAND._serialized_end=3657
  _ENDIANNESS._serialized_start=3659
  _ENDIANNESS._serialized_end=3692
  _TXPOWER._serialized_start=3694
  _TXPOWER._serialized_end=3734
  _JAMMINGMODE._serialized_start=3736
  _JAMMINGMODE._serialized_end=3779
  _LORASPREADINGFACTOR._serialized_start=3781
  _LORASPREADINGFACTOR._serialized_end=3859
  _LORACODINGRATE._serialized_start=3861
  _LORACODINGRATE._serialized_end=3917
  _MODULATION._serialized_start=3919
  _MODULATION._serialized_end=4011
  _SETASKMODULATIONCMD._serialized_start=36
  _SETASKMODULATIONCMD._serialized_end=70
  _SETFSKMODULATIONCMD._serialized_start=72
  _SETFSKMODULATIONCMD._serialized_end=112
  _SET4FSKMODULATIONCMD._serialized_start=114
  _SET4FSKMODULATIONCMD._serialized_end=155
  _SETGFSKMODULATIONCMD._serialized_start=157
  _SETGFSKMODULATIONCMD._serialized_end=198
  _SETMSKMODULATIONCMD._serialized_start=200
  _SETMSKMODULATIONCMD._serialized_end=240
  _SETBPSKMODULATIONCMD._serialized_start=242
  _SETBPSKMODULATIONCMD._serialized_end=264
  _SETQPSKMODULATIONCMD._serialized_start=266
  _SETQPSKMODULATIONCMD._serialized_end=309
  _SETLORAMODULATIONCMD._serialized_start=312
  _SETLORAMODULATIONCMD._serialized_end=534
  _GETSUPPORTEDFREQUENCIESCMD._serialized_start=536
  _GETSUPPORTEDFREQUENCIESCMD._serialized_end=564
  _SETFREQUENCYCMD._serialized_start=566
  _SETFREQUENCYCMD._serialized_end=602
  _SETDATARATECMD._serialized_start=604
  _SETDATARATECMD._serialized_end=634
  _SETENDIANNESSCMD._serialized_start=636
  _SETENDIANNESSCMD._serialized_end=691
  _SETTXPOWERCMD._serialized_start=693
  _SETTXPOWERCMD._serialized_end=740
  _SETPACKETSIZECMD._serialized_start=742
  _SETPACKETSIZECMD._serialized_end=781
  _SETSYNCWORDCMD._serialized_start=783
  _SETSYNCWORDCMD._serialized_end=818
  _SNIFFCMD._serialized_start=820
  _SNIFFCMD._serialized_end=868
  _SENDCMD._serialized_start=870
  _SENDCMD._serialized_end=895
  _SENDRAWCMD._serialized_start=897
  _SENDRAWCMD._serialized_end=925
  _SCHEDULESENDCMD._serialized_start=927
  _SCHEDULESENDCMD._serialized_end=979
  _SCHEDULEPACKETRESP._serialized_start=981
  _SCHEDULEPACKETRESP._serialized_end=1041
  _SCHEDULEPACKETSENT._serialized_start=1043
  _SCHEDULEPACKETSENT._serialized_end=1075
  _STARTCMD._serialized_start=1077
  _STARTCMD._serialized_end=1087
  _STOPCMD._serialized_start=1089
  _STOPCMD._serialized_end=1098
  _JAMCMD._serialized_start=1100
  _JAMCMD._serialized_end=1140
  _MONITORCMD._serialized_start=1142
  _MONITORCMD._serialized_end=1154
  _PACKETRECEIVED._serialized_start=1157
  _PACKETRECEIVED._serialized_end=1399
  _RAWPACKETRECEIVED._serialized_start=1402
  _RAWPACKETRECEIVED._serialized_end=1663
  _PACKETSENT._serialized_start=1665
  _PACKETSENT._serialized_end=1696
  _JAMMED._serialized_start=1698
  _JAMMED._serialized_end=1725
  _MONITORINGREPORT._serialized_start=1727
  _MONITORINGREPORT._serialized_end=1784
  _SUPPORTEDFREQUENCYRANGES._serialized_start=1787
  _SUPPORTEDFREQUENCYRANGES._serialized_end=1931
  _SUPPORTEDFREQUENCYRANGES_FREQUENCYRANGE._serialized_start=1887
  _SUPPORTEDFREQUENCYRANGES_FREQUENCYRANGE._serialized_end=1931
  _MESSAGE._serialized_start=1934
  _MESSAGE._serialized_end=3231
# @@protoc_insertion_point(module_scope)
