# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/unifying/unifying.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%whad/protocol/unifying/unifying.proto\x12\x08unifying\"$\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"K\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0c\x12\x1d\n\x15show_acknowledgements\x18\x03 \x01(\x08\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"E\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"H\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"(\n\x15LogitechDongleModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"*\n\x17LogitechKeyboardModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\'\n\x14LogitechMouseModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"\x11\n\x0fSniffPairingCmd\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\xbe\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xbb\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xea\x04\n\x07Message\x12\x34\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x1b.unifying.SetNodeAddressCmdH\x00\x12#\n\x05sniff\x18\x02 \x01(\x0b\x32\x12.unifying.SniffCmdH\x00\x12\x1f\n\x03jam\x18\x03 \x01(\x0b\x32\x10.unifying.JamCmdH\x00\x12!\n\x04send\x18\x04 \x01(\x0b\x32\x11.unifying.SendCmdH\x00\x12(\n\x08send_raw\x18\x05 \x01(\x0b\x32\x14.unifying.SendRawCmdH\x00\x12\x31\n\x06\x64ongle\x18\x06 \x01(\x0b\x32\x1f.unifying.LogitechDongleModeCmdH\x00\x12\x35\n\x08keyboard\x18\x07 \x01(\x0b\x32!.unifying.LogitechKeyboardModeCmdH\x00\x12\x32\n\x05mouse\x18\x08 \x01(\x0b\x32!.unifying.LogitechKeyboardModeCmdH\x00\x12#\n\x05start\x18\t \x01(\x0b\x32\x12.unifying.StartCmdH\x00\x12!\n\x04stop\x18\n \x01(\x0b\x32\x11.unifying.StopCmdH\x00\x12\x32\n\rsniff_pairing\x18\x0e \x01(\x0b\x32\x19.unifying.SniffPairingCmdH\x00\x12\"\n\x06jammed\x18\x0b \x01(\x0b\x32\x10.unifying.JammedH\x00\x12+\n\x07raw_pdu\x18\x0c \x01(\x0b\x32\x18.unifying.RawPduReceivedH\x00\x12$\n\x03pdu\x18\r \x01(\x0b\x32\x15.unifying.PduReceivedH\x00\x42\x05\n\x03msg*\xc0\x01\n\x0fUnifyingCommand\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x08\n\x04Send\x10\x03\x12\x0b\n\x07SendRaw\x10\x04\x12\x16\n\x12LogitechDongleMode\x10\x05\x12\x18\n\x14LogitechKeyboardMode\x10\x06\x12\x15\n\x11LogitechMouseMode\x10\x07\x12\t\n\x05Start\x10\x08\x12\x08\n\x04Stop\x10\t\x12\x10\n\x0cSniffPairing\x10\nb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.unifying.unifying_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _UNIFYINGCOMMAND._serialized_start=1541
  _UNIFYINGCOMMAND._serialized_end=1733
  _SETNODEADDRESSCMD._serialized_start=51
  _SETNODEADDRESSCMD._serialized_end=87
  _SNIFFCMD._serialized_start=89
  _SNIFFCMD._serialized_end=164
  _JAMCMD._serialized_start=166
  _JAMCMD._serialized_end=191
  _SENDCMD._serialized_start=193
  _SENDCMD._serialized_end=262
  _SENDRAWCMD._serialized_start=264
  _SENDRAWCMD._serialized_end=336
  _LOGITECHDONGLEMODECMD._serialized_start=338
  _LOGITECHDONGLEMODECMD._serialized_end=378
  _LOGITECHKEYBOARDMODECMD._serialized_start=380
  _LOGITECHKEYBOARDMODECMD._serialized_end=422
  _LOGITECHMOUSEMODECMD._serialized_start=424
  _LOGITECHMOUSEMODECMD._serialized_end=463
  _STARTCMD._serialized_start=465
  _STARTCMD._serialized_end=475
  _STOPCMD._serialized_start=477
  _STOPCMD._serialized_end=486
  _SNIFFPAIRINGCMD._serialized_start=488
  _SNIFFPAIRINGCMD._serialized_end=505
  _JAMMED._serialized_start=507
  _JAMMED._serialized_end=534
  _RAWPDURECEIVED._serialized_start=537
  _RAWPDURECEIVED._serialized_end=727
  _PDURECEIVED._serialized_start=730
  _PDURECEIVED._serialized_end=917
  _MESSAGE._serialized_start=920
  _MESSAGE._serialized_end=1538
# @@protoc_insertion_point(module_scope)
