# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/esb/esb.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/esb/esb.proto\x12\x03\x65sb\"$\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"K\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0c\x12\x1d\n\x15show_acknowledgements\x18\x03 \x01(\x08\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"E\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"H\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\")\n\x16PrimaryReceiverModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\",\n\x19PrimaryTransmitterModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\xbe\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xbb\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xc1\x03\n\x07Message\x12/\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x16.esb.SetNodeAddressCmdH\x00\x12\x1e\n\x05sniff\x18\x02 \x01(\x0b\x32\r.esb.SniffCmdH\x00\x12\x1a\n\x03jam\x18\x03 \x01(\x0b\x32\x0b.esb.JamCmdH\x00\x12\x1c\n\x04send\x18\x04 \x01(\x0b\x32\x0c.esb.SendCmdH\x00\x12#\n\x08send_raw\x18\x05 \x01(\x0b\x32\x0f.esb.SendRawCmdH\x00\x12*\n\x03prx\x18\x06 \x01(\x0b\x32\x1b.esb.PrimaryReceiverModeCmdH\x00\x12-\n\x03ptx\x18\x07 \x01(\x0b\x32\x1e.esb.PrimaryTransmitterModeCmdH\x00\x12\x1e\n\x05start\x18\x08 \x01(\x0b\x32\r.esb.StartCmdH\x00\x12\x1c\n\x04stop\x18\t \x01(\x0b\x32\x0c.esb.StopCmdH\x00\x12\x1d\n\x06jammed\x18\n \x01(\x0b\x32\x0b.esb.JammedH\x00\x12&\n\x07raw_pdu\x18\x0b \x01(\x0b\x32\x13.esb.RawPduReceivedH\x00\x12\x1f\n\x03pdu\x18\x0c \x01(\x0b\x32\x10.esb.PduReceivedH\x00\x42\x05\n\x03msg*\x95\x01\n\nESBCommand\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x08\n\x04Send\x10\x03\x12\x0b\n\x07SendRaw\x10\x04\x12\x17\n\x13PrimaryReceiverMode\x10\x05\x12\x1a\n\x16PrimaryTransmitterMode\x10\x06\x12\t\n\x05Start\x10\x07\x12\x08\n\x04Stop\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.esb.esb_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ESBCOMMAND._serialized_start=1300
  _ESBCOMMAND._serialized_end=1449
  _SETNODEADDRESSCMD._serialized_start=36
  _SETNODEADDRESSCMD._serialized_end=72
  _SNIFFCMD._serialized_start=74
  _SNIFFCMD._serialized_end=149
  _JAMCMD._serialized_start=151
  _JAMCMD._serialized_end=176
  _SENDCMD._serialized_start=178
  _SENDCMD._serialized_end=247
  _SENDRAWCMD._serialized_start=249
  _SENDRAWCMD._serialized_end=321
  _PRIMARYRECEIVERMODECMD._serialized_start=323
  _PRIMARYRECEIVERMODECMD._serialized_end=364
  _PRIMARYTRANSMITTERMODECMD._serialized_start=366
  _PRIMARYTRANSMITTERMODECMD._serialized_end=410
  _STARTCMD._serialized_start=412
  _STARTCMD._serialized_end=422
  _STOPCMD._serialized_start=424
  _STOPCMD._serialized_end=433
  _JAMMED._serialized_start=435
  _JAMMED._serialized_end=462
  _RAWPDURECEIVED._serialized_start=465
  _RAWPDURECEIVED._serialized_end=655
  _PDURECEIVED._serialized_start=658
  _PDURECEIVED._serialized_end=845
  _MESSAGE._serialized_start=848
  _MESSAGE._serialized_end=1297
# @@protoc_insertion_point(module_scope)
