from whad.zigbee.crypto import TouchlinkKeyManager, TransportKeyDistribution


analyzers = {
    "touchlink_key_cracking" : TouchlinkKeyManager,
    "transport_key_cracking" : TransportKeyDistribution
}
